import requests
import json
from typing import Optional
from dbt.context.exceptions_jinja import raise_compiler_error
from requests.adapters import HTTPAdapter
from urllib3 import Retry

IOMETE_DEFAULT_CATALOG_NAME = "spark_catalog"

class SchemaService:
    def __init__(self, credentials):
        self.credentials = credentials

        adapter = HTTPAdapter(max_retries=Retry(total=3, backoff_factor=0.5, allowed_methods=None,
                                                status_forcelist=[429, 500, 502, 503, 504]))
        self.session = requests.Session()
        self.session.mount(credentials.scheme, adapter)

    def get_tables_by_namespace(self, namespace: str) -> list:
        final_namespace = namespace.split(".").pop()

        return self._get_namespaces(
            path=f"{final_namespace}/tables?includeMetadata=true",
            error_message=f"Could not get tables for schema {final_namespace}") or []

    def get_table(self, namespace: str, table_name: str) -> Optional[dict]:
        final_namespace = namespace.split(".").pop()

        return self._get_namespaces(
            path=f"{final_namespace}/tables/{table_name}",
            error_message=f"Could not get table metadata for {final_namespace}.{table_name}")

    def _get_namespaces(self, path: str, error_message: str):
        try:
            catalog = self.credentials.schema.split(".")[0] if "." in self.credentials.schema else IOMETE_DEFAULT_CATALOG_NAME
            namespaces = f"{self.credentials.scheme}://{self.credentials.host}:{self.credentials.port}/api/v1/domains/{self.credentials.domain}/schema/catalogs/{catalog}/namespaces"

            response = self.session.get(f"{namespaces}/{path}", timeout=10,
                                        headers={"X-API-TOKEN": self.credentials.token})
            if response.status_code == 404:
                return None

            response.raise_for_status()
            return json.loads(response.text)
        except requests.exceptions.HTTPError as err:
            raise_compiler_error(
                f"{error_message}. "
                f"Request failed with status: {err.response.status_code} and error message is: {err.response.text}"
            )
        except requests.exceptions.RequestException as e:
            raise_compiler_error(f"{error_message}. Request failed with error: {e}")



import json
from concurrent.futures import Future
from dataclasses import dataclass
from typing import Optional, List, Dict, Any, Union, Iterable, Type
import agate
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.relation import RelationType
from dbt.context.exceptions_jinja import raise_compiler_error

import dbt
import dbt.exceptions

from dbt.adapters.base import AdapterConfig, PythonJobHelper
from dbt.adapters.base.impl import catch_as_completed
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.iomete import SparkConnectionManager
from dbt.adapters.iomete import SparkRelation
from dbt.adapters.iomete import SparkColumn
from dbt.adapters.base import BaseRelation
from dbt.clients.agate_helper import DEFAULT_TYPE_TESTER
from dbt_common.utils import AttrDict
from dbt.events import AdapterLogger
from dbt.utils import executor
import sentry_sdk

from dbt.adapters.iomete.python_job import IometeSparkJobHelper
from dbt.adapters.iomete.schema_service import SchemaService, IOMETE_DEFAULT_CATALOG_NAME

logger = AdapterLogger("iomete")

LIST_SCHEMAS_MACRO_NAME = 'list_schemas'
FETCH_TBL_PROPERTIES_MACRO_NAME = 'fetch_tbl_properties'
GET_COLUMNS_IN_RELATION_RAW_MACRO_NAME = "get_columns_in_relation_raw"

KEY_TABLE_OWNER = 'Owner'
KEY_TABLE_STATISTICS = 'Statistics'

sentry_sdk.init(
    dsn="https://a1424d21130340e4913bd8bc1b228c12@o1140336.ingest.sentry.io/4504214031695872",

    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for performance monitoring.
    # We recommend adjusting this value in production.
    traces_sample_rate=1.0,
    attach_stacktrace=True
)


@dataclass
class SparkConfig(AdapterConfig):
    file_format: str = 'parquet'
    location_root: Optional[str] = None
    partition_by: Optional[Union[List[str], str]] = None
    clustered_by: Optional[Union[List[str], str]] = None
    buckets: Optional[int] = None
    options: Optional[Dict[str, str]] = None
    merge_update_columns: Optional[str] = None


class SparkAdapter(SQLAdapter):
    Relation = SparkRelation
    Column = SparkColumn
    ConnectionManager = SparkConnectionManager
    AdapterSpecificConfigs = SparkConfig

    def __init__(self, config):
        super().__init__(config)
        self.schema_service = SchemaService(credentials=config.credentials)

    @classmethod
    def date_function(cls) -> str:
        return 'current_timestamp()'

    @classmethod
    def convert_text_type(cls, agate_table, col_idx):
        return "string"

    @classmethod
    def convert_number_type(cls, agate_table, col_idx):
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "double" if decimals else "bigint"

    @classmethod
    def convert_date_type(cls, agate_table, col_idx):
        return "date"

    @classmethod
    def convert_time_type(cls, agate_table, col_idx):
        return "time"

    @classmethod
    def convert_datetime_type(cls, agate_table, col_idx):
        return "timestamp"

    def quote(self, identifier):
        return '`{}`'.format(identifier)

    def list_relations_without_caching(
            self, schema_relation: SparkRelation
    ) -> List[SparkRelation]:
        tables = self.schema_service.get_tables_by_namespace(str(schema_relation))

        relations = []
        for table in tables:
            rel_type = RelationType.Table
            if table['type'] and table['type'].lower() == 'view':
                rel_type = RelationType.View

            provider = table['provider'].lower() if table['provider'] else None
            schema = table['namespace'] if table['catalog'] == IOMETE_DEFAULT_CATALOG_NAME else f"{table['catalog']}.{table['namespace']}"

            relation = self.Relation.create(
                schema=schema,
                identifier=table['name'],
                type=rel_type,
                provider=provider,
                is_iceberg=provider == "iceberg",
                table_fields=table["columns"],
            )
            relations.append(relation)

        return relations

    def get_relation(
            self, database: str, schema: str, identifier: str
    ) -> Optional[BaseRelation]:
        if not self.Relation.get_default_include_policy().database:
            database = None  # type: ignore

        return super().get_relation(database, schema, identifier)

    def get_columns_in_relation(self, relation: Relation) -> List[SparkColumn]:
        is_temp_table = relation.schema is None and relation.identifier.endswith("tmp")
        if is_temp_table:
            return self._get_columns_of_temp_table(relation)

        rows: AttrDict = self.execute_macro(
            GET_COLUMNS_IN_RELATION_RAW_MACRO_NAME, kwargs={"relation": relation}
        )
        columns = self.parse_describe_extended(relation, rows)
        return columns

    def _get_columns_of_temp_table(self, relation: Relation) -> List[SparkColumn]:
        try:
            desc_table_columns = self.execute_macro(
                'describe_temp_view',
                kwargs={"relation": relation}
            )

            return [SparkColumn(
                table_database=None,
                table_schema=relation.schema,
                table_name=relation.name,
                table_type=relation.type,
                table_owner=None,
                table_stats=None,
                column=column['col_name'],
                column_index=idx,
                dtype=column['data_type'],
            ) for idx, column in enumerate(desc_table_columns)]
        except dbt.exceptions.DbtRuntimeError as e:
            # spark would throw error when table doesn't exist, where other
            # CDW would just return and empty list, normalizing the behavior here
            errmsg = getattr(e, "msg", "")
            if "Table or view not found" in errmsg or "NoSuchTableException" in errmsg:
                return []
            else:
                raise e

    def _get_columns_for_catalog(
            self, relation: SparkRelation
    ) -> Iterable[Dict[str, Any]]:
        logger.warning("_get_columns_for_catalog {}", relation.__dict__)
        columns = self.get_columns_in_relation(relation)

        for column in columns:
            # convert SparkColumns into catalog dicts
            as_dict = column.to_column_dict()
            as_dict['column_name'] = as_dict.pop('column', None)
            as_dict['column_type'] = as_dict.pop('dtype')
            as_dict['table_database'] = None
            yield as_dict

    def get_catalog(self, manifest):

        schema_map = self._get_catalog_schemas(manifest)
        if len(schema_map) > 1:
            raise_compiler_error(
                f'Expected only one database in get_catalog, found '
                f'{list(schema_map)}'
            )

        logger.warning("get_catalog1 {}", schema_map)
        logger.warning("get_catalog2 {}", self.config)

        with executor(self.config) as tpe:
            futures: List[Future[agate.Table]] = []
            for info, schemas in schema_map.items():
                for schema in schemas:
                    futures.append(tpe.submit_connected(
                        self, schema,
                        self._get_one_catalog, info, [schema], manifest
                    ))
            catalogs, exceptions = catch_as_completed(futures)
        return catalogs, exceptions

    def _get_one_catalog(
            self, information_schema, schemas, manifest,
    ) -> agate.Table:
        logger.warning("_get_one_catalog1 {}", information_schema.__dict__)
        logger.warning("_get_one_catalog2 {}", schemas)
        if len(schemas) != 1:
            raise_compiler_error(
                f'Expected only one schema in spark _get_one_catalog, found '
                f'{schemas}'
            )

        database = information_schema.database
        schema = list(schemas)[0]

        columns: List[Dict[str, Any]] = []
        for relation in self.list_relations(database, schema):
            logger.debug("Getting table schema for relation {}", relation)
            columns.extend(self._get_columns_for_catalog(relation))
        return agate.Table.from_object(
            columns, column_types=DEFAULT_TYPE_TESTER
        )

    def check_schema_exists(self, database, schema):
        results = self.execute_macro(
            LIST_SCHEMAS_MACRO_NAME,
            kwargs={'database': database}
        )

        exists = True if schema in [row[0] for row in results] else False
        return exists

    def get_rows_different_sql(
            self,
            relation_a: BaseRelation,
            relation_b: BaseRelation,
            column_names: Optional[List[str]] = None,
            except_operator: str = 'EXCEPT',
    ) -> str:
        """Generate SQL for a query that returns a single row with a two
        columns: the number of rows that are different between the two
        relations and the number of mismatched rows.
        """
        # This method only really exists for test reasons.
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((self.quote(c.name) for c in columns))
        else:
            names = sorted((self.quote(n) for n in column_names))
        columns_csv = ', '.join(names)

        sql = COLUMNS_EQUAL_SQL.format(
            columns=columns_csv,
            relation_a=str(relation_a),
            relation_b=str(relation_b),
        )

        return sql

    def parse_describe_extended(
        self, relation: BaseRelation, raw_rows: AttrDict
    ) -> List[SparkColumn]:
        # Convert the Row to a dict
        dict_rows = [dict(zip(row._keys, row._values)) for row in raw_rows]
        # Find the separator between the rows and the metadata provided
        # by the DESCRIBE TABLE EXTENDED statement
        pos = self.find_table_information_separator(dict_rows)

        # Remove rows that start with a hash, they are comments
        rows = [row for row in raw_rows[0:pos] if not row["col_name"].startswith("#")]
        return [
            SparkColumn(
                table_database=None,
                table_schema=relation.schema,
                table_name=relation.name,
                table_type=relation.type,
                table_owner=None,
                table_stats=None,
                column=column["col_name"],
                column_index=idx,
                dtype=column["data_type"],
            )
            for idx, column in enumerate(rows)
        ]

    @property
    def python_submission_helpers(self) -> Dict[str, Type[PythonJobHelper]]:
        return {
            "spark_job": IometeSparkJobHelper,
        }

    @property
    def default_python_submission_method(self) -> str:
        return "spark_job"

    def generate_python_submission_response(self, submission_result: Any) -> AdapterResponse:
        return AdapterResponse(_message="OK")

    # In original implementation handle.commit() and handle.rollback() are called.
    # Since spark doesn't support transactions, this implementation is a no-op.
    def run_sql_for_tests(self, sql, fetch, conn):
        cursor = conn.handle.cursor()
        try:
            cursor.execute(sql)
            if fetch == "one":
                if hasattr(cursor, "fetchone"):
                    return cursor.fetchone()
                else:
                    # AttributeError: 'PyhiveConnectionWrapper' object has no attribute 'fetchone'
                    return cursor.fetchall()[0]
            elif fetch == "all":
                return cursor.fetchall()
            else:
                return
        except BaseException as e:
            print(sql)
            print(e)
            raise
        finally:
            conn.transaction_open = False

    @staticmethod
    def find_table_information_separator(rows: List[dict]) -> int:
        pos = 0
        for row in rows:
            if not row["col_name"] or row["col_name"].startswith("#"):
                break
            pos += 1
        return pos


# spark does something interesting with joins when both tables have the same
# static values for the join condition and complains that the join condition is
# "trivial". Which is true, though it seems like an unreasonable cause for
# failure! It also doesn't like the `from foo, bar` syntax as opposed to
# `from foo cross join bar`.
COLUMNS_EQUAL_SQL = '''
with diff_count as (
    SELECT
        1 as id,
        COUNT(*) as num_missing FROM (
            (SELECT {columns} FROM {relation_a} EXCEPT
             SELECT {columns} FROM {relation_b})
             UNION ALL
            (SELECT {columns} FROM {relation_b} EXCEPT
             SELECT {columns} FROM {relation_a})
        ) as a
), table_a as (
    SELECT COUNT(*) as num_rows FROM {relation_a}
), table_b as (
    SELECT COUNT(*) as num_rows FROM {relation_b}
), row_count_diff as (
    select
        1 as id,
        table_a.num_rows - table_b.num_rows as difference
    from table_a
    cross join table_b
)
select
    row_count_diff.difference as row_count_difference,
    diff_count.num_missing as num_mismatched
from row_count_diff
cross join diff_count
'''.strip()



from dbt.adapters.iomete.connections import SparkConnectionManager  # noqa
from dbt.adapters.iomete.connections import SparkCredentials  # noqa
from dbt.adapters.iomete.relation import SparkRelation  # noqa
from dbt.adapters.iomete.column import SparkColumn  # noqa
from dbt.adapters.iomete.impl import SparkAdapter

from dbt.adapters.base import AdapterPlugin
from dbt.include import iomete

Plugin = AdapterPlugin(
    adapter=SparkAdapter,
    credentials=SparkCredentials,
    include_path=iomete.PACKAGE_PATH
)



import random
import string
import time
from enum import Enum

from typing import Dict, Any
import uuid

import dbt.exceptions
from dbt.adapters.base import PythonJobHelper
from dbt.adapters.iomete import SparkCredentials
from dbt.events import AdapterLogger

from iomete_sdk.spark import SparkJobApiClient

POLLING_PERIOD_SECONDS = 10
DEFAULT_TIMEOUT = 60 * 60 * 24

logger = AdapterLogger("iomete")


class IometeSparkJobHelper(PythonJobHelper):
    def __init__(self, parsed_model: Dict, credential: SparkCredentials) -> None:
        self.parsed_model = parsed_model
        self.alias = parsed_model["alias"]
        self.schema = parsed_model["schema"]
        self.protocol = 'https://' if credential.https else 'http://'
        self.iom_client = SparkJobApiClient(
            host=f"{self.protocol}{credential.host}:{credential.port}",
            api_key=credential.token,
            domain=credential.domain
        )

    @property
    def job_id(self) -> str:
        spark_job_id = self.parsed_model["config"].get("spark_job_id")
        if not spark_job_id:
            raise ValueError("spark_job_id is required for submitting python models.")
        return spark_job_id

    def submit(self, compiled_code: str) -> Any:
        config_overrides = self.parsed_model["config"].get("spark_job_overrides", {})

        run_response = self._submit_job_run(job_id=self.job_id, payload={
            "pythonScript": compiled_code,
            "arguments": config_overrides.get("arguments", None),
            "envVars": config_overrides.get("envVars", None),
            "sparkConf": config_overrides.get("sparkConf", None),
        })
        logger.info(f"Spark job ({self.job_id}) triggered with run_id: {run_response['id']}")

        self._monitor_state(job_id=self.job_id, run_id=run_response["id"])

    def _create_job(self, payload):
        response = self.iom_client.create_job(payload=payload)
        return response

    def _submit_job_run(self, job_id, payload):
        response = self.iom_client.submit_job_run(job_id=job_id, payload=payload)
        return response

    def _get_job_run(self, job_id, run_id):
        response = self.iom_client.get_job_run_by_id(job_id=job_id, run_id=run_id)
        return response

    def _monitor_state(self, job_id, run_id):
        logger.info(f"Spark job submitted with job_id: {job_id}")

        while True:
            app = self._get_job_run(job_id, run_id)
            app_state = _get_state_from_app(app)
            if app_state.is_final:
                if app_state.is_successful:
                    logger.info(f"{job_id} completed successfully.")
                    return
                else:
                    error_message = "Job {j} failed with terminal state: {s}".format(
                        j=job_id, s=app_state.value
                    )
                    raise dbt.exceptions.DbtRuntimeError(error_message)
            else:
                logger.info("{} in app state: {}", job_id, app_state.value)
                logger.info("Sleeping for {} seconds.", POLLING_PERIOD_SECONDS)
                time.sleep(POLLING_PERIOD_SECONDS)


def _get_state_from_app(app):
    return ApplicationStateType(app.get("driverStatus", ""))


class ApplicationStateType(Enum):
    EmptyState = "ENQUEUED"
    DeployingState = "SUBMITTED"
    RunningState = "RUNNING"
    CompletedState = "COMPLETED"
    FailedState = "FAILED"
    AbortedState = "ABORTED"
    AbortingState = "ABORTING"
    ExecutorState = map(
        {"RUNNING": 1},
        {"PENDING": 1},
    )

    @property
    def is_final(self) -> bool:
        return self in [
            ApplicationStateType.CompletedState,
            ApplicationStateType.FailedState,
            ApplicationStateType.AbortedState,
        ]

    @property
    def is_successful(self) -> bool:
        return self == ApplicationStateType.CompletedState
